@@ -1,4 +1,4 @@
-#     Copyright 2024, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file
+#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file
 
 
 """ Standard plug-in to tell Nuitka about implicit imports.
@@ -15,21 +15,20 @@ import os
 from nuitka.__past__ import iter_modules, unicode
 from nuitka.importing.Importing import locateModule
 from nuitka.importing.Recursion import decideRecursion
-from nuitka.plugins.PluginBase import NuitkaPluginBase
+from nuitka.plugins.YamlPluginBase import NuitkaYamlPluginBase
 from nuitka.utils.ModuleNames import ModuleName
 from nuitka.utils.Utils import isMacOS, isWin32Windows
-from nuitka.utils.Yaml import getYamlPackageConfiguration
 
 
-class NuitkaPluginImplicitImports(NuitkaPluginBase):
+class NuitkaPluginImplicitImports(NuitkaYamlPluginBase):
     plugin_name = "implicit-imports"
-
     plugin_desc = (
         "Provide implicit imports of package as per package configuration files."
     )
+    plugin_category = "core"
 
     def __init__(self):
-        self.config = getYamlPackageConfiguration()
+        NuitkaYamlPluginBase.__init__(self)
 
         self.lazy_loader_usages = {}
 
@@ -85,6 +84,32 @@ class NuitkaPluginImplicitImports(NuitkaPluginBase):
 
         yield current
 
+    def _resolveImplicitImportsConfig(self, full_name, dependency):
+        if "(" in dependency:
+            value = self.evaluateExpression(
+                config_module_name=full_name,
+                module_name=full_name,
+                expression=dependency,
+                config_name="depends value",
+                extra_context=None,
+                single_value=False,
+            )
+
+            if type(value) in (str, unicode):
+                value = (value,)
+
+            for v in value:
+                if "*" in v or "?" in v:
+                    for resolved in self._resolveModulePattern(v):
+                        yield resolved
+                else:
+                    yield ModuleName(v)
+        elif "*" in dependency or "?" in dependency:
+            for resolved in self._resolveModulePattern(dependency):
+                yield resolved
+        else:
+            yield dependency
+
     def _handleImplicitImportsConfig(self, module, config):
         full_name = module.getFullName()
 
@@ -97,31 +122,17 @@ class NuitkaPluginImplicitImports(NuitkaPluginBase):
                     dependency = full_name.getChildNamed(dependency[1:]).asString()
                 elif full_name.getPackageName() is None:
                     # Not a package, potentially a naming conflict, when
-                    # compiling with "--module" something that matches a PyPI
-                    # name.
+                    # compiling with "--mode=module" something that matches a
+                    # PyPI name.
                     continue
                 else:
                     dependency = full_name.getSiblingNamed(dependency[1:]).asString()
 
-            if "(" in dependency:
-                value = self.evaluateExpression(
-                    full_name=full_name,
-                    expression=dependency,
-                    config_name="depends value",
-                    extra_context=None,
-                    single_value=False,
-                )
-
-                if type(value) in (str, unicode):
-                    value = (value,)
-
-                for v in value:
-                    yield v
-            elif "*" in dependency or "?" in dependency:
-                for resolved in self._resolveModulePattern(dependency):
-                    yield resolved
-            else:
-                yield dependency
+            for value in self._resolveImplicitImportsConfig(
+                full_name=full_name,
+                dependency=dependency,
+            ):
+                yield value
 
     def _getImportsByFullname(self, module, full_name):
         """Provides names of modules to imported implicitly."""
@@ -256,6 +267,9 @@ class NuitkaPluginImplicitImports(NuitkaPluginBase):
             elif full_name == crypto_module_name + ".Cipher.PKCS1_v1_5":
                 yield crypto_module_name + ".Cipher._pkcs1_decode"
 
+            elif full_name == crypto_module_name + ".Cipher.PKCS1_OAEP":
+                yield crypto_module_name + ".Cipher._pkcs1_decode"
+
             elif full_name == crypto_module_name + ".Math._IntegerCustom":
                 yield crypto_module_name + ".Math._modexp"
 
@@ -304,8 +318,6 @@ class NuitkaPluginImplicitImports(NuitkaPluginBase):
         for config_package_dir in config.get("package-dirs", ()):
             yield os.path.normpath(os.path.join(package_dir, "..", config_package_dir))
 
-            yield package_dir
-
         for config_package_name in config.get("package-paths", ()):
             module_filename = self.locateModule(config_package_name)
 
@@ -348,19 +360,9 @@ class NuitkaPluginImplicitImports(NuitkaPluginBase):
                     yield item
 
     def onModuleSourceCode(self, module_name, source_filename, source_code):
-        # Too much code here, pylint: disable=too-many-branches
-        # TODO: Move the ones that would be possible to yaml config,
-        # e.g. the numexpr hack.
-
-        if module_name == "numexpr.cpuinfo":
-            # We cannot intercept "is" tests, but need it to be "isinstance",
-            # so we patch it on the file. TODO: This is only temporary, in
-            # the future, we may use optimization that understands the right
-            # hand size of the "is" argument well enough to allow for our
-            # type too.
-            source_code = source_code.replace(
-                "type(attr) is types.MethodType", "isinstance(attr, types.MethodType)"
-            )
+        # TODO: Move the ones that would be possible to yaml config, or a
+        # separate plugin for lazy loaders.
+        # pylint: disable=too-many-branches,too-many-locals
 
         if module_name == "site":
             if source_code.startswith("def ") or source_code.startswith("class "):
@@ -417,9 +419,9 @@ __file__ = (__nuitka_binary_dir + '%ssite.py') if '__nuitka_binary_dir' in dict(
 
         if module_name == "huggingface_hub":
             # Special handling for huggingface that uses the source code variant
-            # of lazy module. spell-checker: ignore huggingface
+            # of lazy module. spell-checker: ignore huggingface,submod
             if (
-                "__getattr__, __dir__, __all__ = _attach(__name__, submodules=[], submod_attrs=_SUBMOD_ATTRS)"
+                "_attach(__name__, submodules=[], submod_attrs=_SUBMOD_ATTRS)"
                 in source_code
             ):
                 huggingface_hub_lazy_loader_info = self.queryRuntimeInformationSingle(
@@ -450,9 +452,17 @@ __file__ = (__nuitka_binary_dir + '%ssite.py') if '__nuitka_binary_dir' in dict(
                 )
 
                 pydantic_lazy_loader_info = {}
+                pydantic_lazy_submodules = []
+
                 for key, value in pydantic_info.items():
                     # Older pydantic had only a string for the attribute.
                     if type(value) is tuple:
+                        # Special case, __module__ means it's a sub-module
+                        if value == ("pydantic", "__module__"):
+                            pydantic_lazy_submodules.append(key)
+                            continue
+
+                        # Otherwise it's a long winded way of specifying a module name.
                         value = "".join(value).rstrip(".")
 
                     if value not in pydantic_lazy_loader_info:
@@ -461,10 +471,48 @@ __file__ = (__nuitka_binary_dir + '%ssite.py') if '__nuitka_binary_dir' in dict(
 
                 self._addLazyLoader(
                     module_name=module_name,
-                    submodules=(),
+                    submodules=pydantic_lazy_submodules,
                     submodule_attrs=pydantic_lazy_loader_info,
                 )
 
+        if module_name == "scipy":
+            # Scipy has its own lazy loading, spell-checker: ignore scipy
+            if "def __getattr__(" in source_code:
+                scipy_info = self.queryRuntimeInformationSingle(
+                    setup_codes="import scipy",
+                    value="scipy.submodules",
+                    info_name="scipy_lazy_loader",
+                )
+
+                self._addLazyLoader(
+                    module_name=module_name,
+                    submodules=scipy_info,
+                    submodule_attrs={},
+                )
+
+        if module_name == "toga":
+            # Toga has lazy loading in some versions.
+
+            if "def __getattr__(" in source_code:
+                toga_info = self.queryRuntimeInformationSingle(
+                    setup_codes="import toga",
+                    value="toga.toga_core_imports",
+                    info_name="toga_lazy_loader",
+                )
+
+                toga_submodule_attrs = {}
+
+                for attribute_name, sub_module_name in toga_info.items():
+                    if sub_module_name not in toga_submodule_attrs:
+                        toga_submodule_attrs[sub_module_name] = []
+                    toga_submodule_attrs[sub_module_name].append(attribute_name)
+
+                self._addLazyLoader(
+                    module_name=module_name,
+                    submodules=(),
+                    submodule_attrs=toga_submodule_attrs,
+                )
+
         return source_code
 
     def _addLazyLoader(self, module_name, submodules, submodule_attrs):
@@ -486,9 +534,11 @@ __file__ = (__nuitka_binary_dir + '%ssite.py') if '__nuitka_binary_dir' in dict(
 
         submodule_attrs = dict(
             (
-                module_name.getChildNamed(submodule[1:])
-                if submodule.startswith(".")
-                else ModuleName(submodule),
+                (
+                    module_name.getChildNamed(submodule[1:])
+                    if submodule.startswith(".")
+                    else ModuleName(submodule)
+                ),
                 tuple(attribute_names),
             )
             for (submodule, attribute_names) in sorted(submodule_attrs.items())
@@ -540,6 +590,10 @@ __file__ = (__nuitka_binary_dir + '%ssite.py') if '__nuitka_binary_dir' in dict(
                 ):
                     code = "\n".join(entry.get("pre-import-code"))
 
+                    # TODO: Add a description to the Yaml file.
+                    yield code, """\
+According to Yaml 'pre-import-code' configuration."""
+
         for entry in self.config.get(full_name, section="import-hacks"):
             if "force-environment-variables" in entry:
                 if self.evaluateCondition(
@@ -652,6 +706,9 @@ According to Yaml 'overridden-environment-variables' configuration."""
         "asyncua.ua.object_ids",  # Too large generated code
         "asyncua.ua.uaerrors._auto",  # Too large generated code
         "asyncua.server.standard_address_space.standard_address_space_services",  # Too large generated code
+        "opcua.ua.object_ids",  # Too large generated code
+        "opcua.ua.uaerrors._auto",  # Too large generated code
+        "opcua.server.server.standard_address_space",
         "azure.mgmt.network",  # Too large generated code
         "azure.mgmt.compute",  # Too large generated code
         "transformers.utils.dummy_pt_objects",  # Not performance relevant.
@@ -659,12 +716,26 @@ According to Yaml 'overridden-environment-variables' configuration."""
         "transformers.utils.dummy_tf_objects",  # Not performance relevant.
         "rich",  #  Not performance relevant and memory leaking due to empty compiled cell leaks
         "altair.vegalite.v5.schema",  # Not performance relevant.
+        "azure",  # Not performance relevant.
+        "networkx",  # Needs solutions for bytecode requiring decorators.
+    )
+
+    unworthy_modulename_patterns = (
+        "tensorflow.*test",  # Not performance relevant.
+        "tensorflow.**.test_util",  # Not performance relevant.
+        "google.protobuf.*_pb2",  #  Too large generated code
     )
 
     def decideCompilation(self, module_name):
         if module_name.hasOneOfNamespaces(self.unworthy_namespaces):
             return "bytecode"
 
+        is_match, _reason = module_name.matchesToShellPatterns(
+            self.unworthy_modulename_patterns
+        )
+        if is_match:
+            return "bytecode"
+
     def onModuleUsageLookAhead(
         self, module_name, module_filename, module_kind, get_module_source
     ):
